grid.board {
  border: none;
  background: #747474;
}

grid.block {
  background: #5b5b5b;
}

.paused {
  background: rgba(0, 0, 0, 0.75);
  font-weight: bold;
  color: #fff;
}

sudokucell {
  background: #454545;
}
sudokucell.highlight-bg {
  background: #717171;
}
sudokucell.selected {
  background: #4c5e79;
}
sudokucell.fixed {
  background: #343434;
}
sudokucell label {
  color: #FFF;
}
sudokucell.selected > label {
  background: #4c5e79;
}
sudokucell.paused {
  background: #454545;
}
sudokucell.paused label {
  color: transparent;
}
sudokucell.highlight-label > label {
  background: #20324b;
}
sudokucell > label.earmark.highlight-label {
  color: #20324b;
  font-weight: bold;
}

.earmark.error,
sudokucell.highlight-label > .earmark.error,
sudokucell.selected > .earmark.error,
sudokucell.error > label {
  color: red;
  font-weight: bold;
}

